﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Dispatcher;
using Microsoft.Practices.EnterpriseLibrary.Common.Configuration;
using Microsoft.Practices.EnterpriseLibrary.PolicyInjection.Configuration;
using Microsoft.Practices.Unity;
using Microsoft.Practices.Unity.Configuration;
using Microsoft.Practices.Unity.InterceptionExtension;
using Artech.InterceptableUnity;

namespace Artech.UnityIntegration
{
    public static class Extension
    {
        public static IList<UnityTypeMapping> Copy(this UnityTypeElementCollection unityTypeElements)
        {
            IList<UnityTypeMapping> mappings = new List<UnityTypeMapping>();
            foreach (UnityTypeElement type in unityTypeElements)
            {
                mappings.Add(new UnityTypeMapping { Type = type.Type, MapTo = type.MapTo, Name = type.Name });
            }

            return mappings;
        }
    }

    public class UnityTypeMapping
    {
        public Type Type
        { get; set; }

        public Type MapTo
        { get; set; }

        public string Name
        { get; set; }
    }


    public class UnityInstanceProvider : IInstanceProvider
    {
        private static object syncHelper = new object();
        private static IDictionary<string, IList<UnityTypeMapping>> registeredTypeMappings;
        public  static IDictionary<string, IUnityContainer> Containers
        { get; private set; }

        private IList<UnityTypeMapping> _registeredTypeMappings;
        public Type ContractType
        { get; private set; }
        public IUnityContainer Container
        { get; private set; }

        static UnityInstanceProvider()
        {
            registeredTypeMappings = new Dictionary<string, IList<UnityTypeMapping>>();
            Containers = new Dictionary<string, IUnityContainer>();
        }

        public UnityInstanceProvider(Type contractType, string containerName)
        {
            if (contractType == null)
            {
                throw new ArgumentNullException("contractType");
            }

            this.ContractType = contractType;

            string key = containerName ?? string.Empty;
            if (Containers.ContainsKey(key))
            {
                this.Container = Containers[key];
                this._registeredTypeMappings = registeredTypeMappings[key];
                return;
            }

            UnityContainerElement containerElement = this.GetUnitySettings(containerName);
            IUnityContainer container = new UnityContainer();
            if (null != containerElement)
            {
                containerElement.Configure(container);
            }
            container.AddNewExtension<ExtendedInterception>();
            PolicyInjectionSettings section = (PolicyInjectionSettings)ConfigurationSourceFactory.Create().GetSection("policyInjection");
            if (section != null)
            {
                section.ConfigureContainer(this.Container, ConfigurationSourceFactory.Create());
            }
            lock (syncHelper)
            {
                if (!Containers.ContainsKey(key))
                {
                    Containers[key] = container;
                    registeredTypeMappings[key] = containerElement.Types.Copy();
                }
            }

            this.Container = container;
            this._registeredTypeMappings = registeredTypeMappings[key];
        }

        #region IInstanceProvider Members

        public object GetInstance(InstanceContext instanceContext, Message message)
        {
            string contractServiceTypeMappingName = string.Empty;
            var contractServiceTypeMappings = from mapping in this._registeredTypeMappings
                                              where mapping.Type == this.ContractType && 
                                              mapping.MapTo == instanceContext.Host.Description.ServiceType
                                              select mapping;
            if (contractServiceTypeMappings.Count() == 0)
            {
                contractServiceTypeMappingName = Guid.NewGuid().ToString();
                this.Container.RegisterType(this.ContractType, instanceContext.Host.Description.ServiceType, contractServiceTypeMappingName);
                this._registeredTypeMappings.Add(new UnityTypeMapping { Type = this.ContractType, MapTo = instanceContext.Host.Description.ServiceType, Name = contractServiceTypeMappingName });
            }
            else
            {
                contractServiceTypeMappingName = contractServiceTypeMappings.ToArray<UnityTypeMapping>()[0].Name;
            }

            return this.Container.Resolve(this.ContractType, contractServiceTypeMappingName);
        }

        private UnityContainerElement GetUnitySettings(string containerName)
        {
            UnityConfigurationSection unitySection = ConfigurationManager.GetSection("unity") as UnityConfigurationSection;
            if (unitySection == null)
            {
                return null;
            }

            if (string.IsNullOrEmpty(containerName))
            {
                return unitySection.Containers.Default;
            }
            else
            {
                return unitySection.Containers[containerName];
            }
        }

        public object GetInstance(InstanceContext instanceContext)
        {
            return this.GetInstance(instanceContext, null);
        }

        public void ReleaseInstance(InstanceContext instanceContext, object instance)
        {
            IDisposable disposable = instance as IDisposable;
            if (disposable != null)
            {
                disposable.Dispose();
            }
        }

        #endregion
    }
}

